package haveubreak;

import haveubreak.msg.ChangementStatut;
import haveubreak.msg.DemandePause;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;
import javax.swing.SwingUtilities;
import org.jgroups.Address;
import org.jgroups.Channel;
import org.jgroups.JChannel;
import org.jgroups.Message;
import org.jgroups.ReceiverAdapter;
import org.jgroups.View;

public class Handler extends ReceiverAdapter {

    private static final Handler INSTANCE = new Handler();
    
    private Configuration config = null;
    private Channel channel = null;
    private final List<HaveubreakListener> listeners = new LinkedList<HaveubreakListener>();
    private final List<GroupMember> members = new ArrayList<GroupMember>();
    private Statut statut = Statut.Travail;
    
    public static Handler getInstance() {
        return INSTANCE;
    }
    
    private Handler() {
    }

    public void nicknameChanged() {
        try {
            send(new ChangementStatut(config.getUserName(), config.getNickName(), statut));
        } catch (Throwable t) {
            t.printStackTrace();
        }
    }
    
    public void setConfig(Configuration config) {
        this.config = config;
    }
    
    public synchronized boolean isConnected() {
        return channel != null;
    }
    
    public synchronized Channel getJChannel() {
        return channel;
    }
    
    public void send(Serializable data) {
        if (!isConnected()) throw new NotConnectedException();
        try {
            channel.send(null, null, data);
        } catch (Throwable t) {
            throw new NotConnectedException(t);
        }
        if (data instanceof String) {
            fireEvent(HaveubreakEvent.MESSAGE_SENT, data);
        }
    }
    
    public void send(Address target, Serializable data) {
        if (!isConnected()) throw new NotConnectedException();
        try {
            channel.send(target, null, data);
        } catch (Throwable t) {
            throw new NotConnectedException(t);
        }
    }
    
    public GroupMember getMemberByAddr(Address src) {
        synchronized (members) {
            for (GroupMember m : members) {
                if (m.getAddr().equals(src)) return m;
            }
        }
        return null;
    }
    
    public Statut getMyCurrentStatut() {
        return statut;
    }
    
    public void setMyCurrentStatut(Statut statut) {
        boolean notify = !this.statut.equals(statut);
        this.statut = statut;
        if (notify) {
            try {
                send(new ChangementStatut(config.getUserName(), config.getNickName(), statut));
            } catch (Throwable t) { }
        }
    }
    
    /** Open/Close **/
    
    public synchronized Exception open() {
        close();
        try {
            
            channel = new JChannel(config.getJgroupsConfig());
            
            channel.setReceiver(this);
            
            channel.connect(config.getChannelName());
            
            send(new ChangementStatut(
                    config.getUserName(),
                    config.getNickName(),
                    getMyCurrentStatut()
            ));
            
        } catch (Exception ex) {
            channel = null;
            return ex;
        }
        
        fireEvent(HaveubreakEvent.CONNECTED, channel.getName());
        
        return null;
    }
    
    public synchronized void close() {
        if (channel == null) return;
        channel.close();
        synchronized (members) {
            members.clear();
        }
        fireEvent(HaveubreakEvent.DISCONNECTED, null);
    }
    
    /** Events **/
    
    public boolean addListener(HaveubreakListener listener) {
        if (listener == null) return false;
        synchronized (listeners) {
            if (listeners.contains(listener)) return false;
            listeners.add(listener);
        }
        return true;
    }
    
    public boolean removeListener(HaveubreakListener listener) {
        if (listener == null) return false;
        synchronized (listeners) {
            return listeners.remove(listener);
        }
    }
    
    public void fireEvent(HaveubreakEvent event, Object msg) {
        synchronized (listeners) {
            for (HaveubreakListener listener : listeners) {
                try {
                    listener.receivedEvent(event, msg);
                } catch (Throwable t) {
                    t.printStackTrace();
                }
            }
        }
    }    
    
    /** JGROUPS Receiver Implementation **/
    
    @Override
    public void receive(Message m) {
        
        super.receive(m);
        
        if (m.getObject() instanceof String) {
            fireEvent(HaveubreakEvent.MESSAGE_RECEIVED, m);
        }
        
        else if (m.getObject() instanceof ChangementStatut) {
            GroupMember member = getMemberByAddr(m.getSrc());
            if (member != null) {
                // Nickname
                String nickName = ((ChangementStatut) m.getObject()).getNickname();
                boolean nickChanged = false;
                if (member.getNickName() == null) {
                    nickChanged = nickName != null;
                } else {
                    nickChanged = !member.getNickName().equals(nickName);
                }
                member.setNickName(nickName);
                // Statut
                Statut statut = ((ChangementStatut) m.getObject()).getStatut();
                boolean statutChanged = false;
                if (member.getStatut() == null) {
                    statutChanged = statut != null;
                } else {
                    statutChanged = !member.getStatut().equals(statut);
                }
                member.setStatut(statut);
                // Username
                member.setUserName(((ChangementStatut) m.getObject()).getName());
                

                if (statutChanged) {
                    fireEvent(HaveubreakEvent.MEMBER_STATUT_CHANGED, member);
                }
                if (nickChanged) {
                    fireEvent(HaveubreakEvent.MEMBER_NICKNAME_CHANGED, member);
                }
            }
        }
        
        else if (m.getObject() instanceof DemandePause) {
            GroupMember member = getMemberByAddr(m.getSrc());
            if (member != null) {
                fireEvent(HaveubreakEvent.MEMBER_DEMANDE_PAUSE, new Object[] {
                    member,
                    (DemandePause) m.getObject()
                });
            }
        }

        else {
            System.err.println("Warning : unkown message from "+m.getSrc()+" : "+m);
        }
        
    }
    
    @Override
    public void viewAccepted(View v) {
        
        super.viewAccepted(v);
        
        List<Address> income = new ArrayList<Address>();
        List<GroupMember> outcome = new ArrayList<GroupMember>();
        
        synchronized (members) {
            for (GroupMember m : members) {
                if (!v.containsMember(m.getAddr())) {
                    outcome.add(m);
                }
            }
            for (Address addr : v.getMembers()) {
                boolean found = false;
                for (GroupMember m : members) {
                    if (m.getAddr().equals(addr)) {
                        found = true;
                        break;
                    }
                }
                if (!found) income.add(addr);
            }
        }
        
        // Lost members
        for (GroupMember m : outcome) {
            synchronized (members) {
                members.remove(m);
            }
            fireEvent(HaveubreakEvent.MEMBER_LOST, m);
        }
        
        // New members
        for (Address a : income) {
            GroupMember m = new GroupMember(a);
            synchronized (members) {
                members.add(m);
            }

            fireEvent(HaveubreakEvent.MEMBER_JOIN, m);
        }

        if (income.size() > 0) {
            SwingUtilities.invokeLater(new Runnable() {
                @Override
                public void run() {
                    try {
                        Thread.sleep(500);
                        send(new ChangementStatut(
                            config.getUserName(),
                            config.getNickName(),
                            getMyCurrentStatut()
                        ));
                    } catch (Throwable t) {
                        t.printStackTrace();
                    }
                }
            });
        }
    }
    
    public boolean isMe(GroupMember m) {
        return isMe(m.getAddr());
    }
    
    public boolean isMe(Address a) {
        if (!isConnected()) return false;
        try {
            return channel.getAddress().equals(a);
        } catch (Throwable t) {
            return false;
        }
    }

}
